#!/bin/bash
SIMULATOR=$1
TELNET_PORT=$2
MAX_PARALLEL_JOBS=$3

if [ -z "${SIMULATOR}" ] || [ -z "${TELNET_PORT}" ]; then
	echo "Usage: $(basename $0) <simulator> <telnet port> [<max parallel jobs>]"
	exit 0
fi

if [ -z "$(nc -h 2>&1 | grep -i 'detach from stdin')" ]; then
	NETCAT_FLAVOR="traditional"
else
	NETCAT_FLAVOR="openbsd"
fi

function Console()
{
	if [ "${NETCAT_FLAVOR}" = "openbsd" ]; then
		nc -d localhost $1
	else
		nc localhost $1
	fi
	echo "Terminated"
}

printf "" > log.txt

BENCHS="\
	telecomm_adpcm_large \
	security_blowfish_large \
	automotive_basicmath_large \
	security_rijndael_large \
	telecomm_gsm_large \
	telecomm_adpcm_small \
	consumer_lame_large \
	telecomm_fft_large \
	network_patricia_large \
	consumer_jpeg_large \
	telecomm_crc32_large \
	automotive_qsort_large \
	security_blowfish_small \
	automotive_susan_large \
	security_rijndael_small \
	consumer_typeset_large \
	automotive_basicmath_small \
	consumer_jpeg_small \
	telecomm_fft_small \
	network_patricia_small \
	automotive_bitcount_large \
	consumer_lame_small \
	network_dijkstra_large \
	security_sha_large \
	automotive_qsort_small \
	consumer_typeset_small \
	automotive_susan_small \
	telecomm_gsm_small \
	telecomm_crc32_small \
	security_pgp \
	office_stringsearch_large \
	network_dijkstra_small \
	automotive_bitcount_small \
	security_sha_small \
	office_stringsearch_small \
"

MAX_PARALLEL_PIDS=1
PID_QUEUE=""
PID_QUEUE_LENGTH=0

function QueuePID()
{
	PID_QUEUE="${PID_QUEUE} $1"
	PID_QUEUE_LENGTH=$((${PID_QUEUE_LENGTH}+1))
}

function DequeuePID()
{
	local OLD_PID_QUEUE=${PID_QUEUE}
	PID_QUEUE=""
	for PID in ${OLD_PID_QUEUE}; do
		if [ ! "${PID}" = "$1" ] ; then
			PID_QUEUE="${PID_QUEUE} ${PID}"
		fi
	done
	PID_QUEUE_LENGTH=$(($PID_QUEUE_LENGTH-1))
}

function CheckPIDQueue()
{
	jobs >/dev/null 2>/dev/null    # strange: this make sure CHILD_PIDS gets updated
	local OLD_PID_QUEUE=${PID_QUEUE}
	CHILD_PIDS=$(jobs -l | sed -n 's/^ *\(\[[0-9]*\]\)\? *\(+\|-\)\? *\([0-9]*\) .*$/\3 /p')
	for PID in ${OLD_PID_QUEUE}; do
		PID_STILL_RUNNING=0
		for CHILD_PID in ${CHILD_PIDS}; do
			if [ ${PID} -eq ${CHILD_PID} ]; then
				PID_STILL_RUNNING=1
				break
			fi
		done
		if [ ${PID_STILL_RUNNING} -eq 0 ]; then
			DequeuePID ${PID}
		fi
	done
}

function WaitUntilReadyToQueuePID()
{
	while [ ${PID_QUEUE_LENGTH} -ge ${MAX_PARALLEL_PIDS} ]; do
		CheckPIDQueue
		sleep 1
	done
}

function Launch()
{
	BENCH="$1"
	TELNET_PORT="$2"
	PROGRESS_TEXT="$3"
	printf "Launching ${BENCH} (${PROGRESS_TEXT})..."
	${SIMULATOR} -c config/sim_config_${BENCH}.xml -s telnet.enable-negotiation=false -s telnet.telnet-tcp-port=${TELNET_PORT} &> sim_out_${BENCH}.txt &   # -s uart-lite.telnet-refresh-time=500us
	PID=$!
	QueuePID ${PID}
	until [ -e "sim_out_${BENCH}.txt" ]; do
		sleep 1
		printf "."
	done
	until [ -n "$(cat sim_out_${BENCH}.txt | grep 'telnet: Listening')" ]; do
		if [ -n "$(cat sim_out_${BENCH}.txt | grep 'ERROR!')" ]; then
			printf "error (see sim_out_${BENCH}.txt)\n"
			return
		fi
		sleep 1
		printf "."
	done
	sleep 1
	Console ${TELNET_PORT} > "console_out_${BENCH}.txt" &
	PID=$!
	QueuePID ${PID}
	printf "ok\n"
}

if [ -z "${MAX_PARALLEL_JOBS}" ]; then
	MAX_PARALLEL_JOBS=1
fi

MAX_PARALLEL_PIDS=$((${MAX_PARALLEL_JOBS} * 2))    # there's one simulator process plus one netcat client process per job

NUM_JOBS=0
NUM_LAUNCHED_JOBS=0

for BENCH in ${BENCHS}; do
	NUM_JOBS=$((${NUM_JOBS} + 1))
done

for BENCH in ${BENCHS}; do
	Launch "${BENCH}" "${TELNET_PORT}" "job #$((${NUM_LAUNCHED_JOBS} + 1)) of ${NUM_JOBS}"
	TELNET_PORT=$((${TELNET_PORT} + 1))
 	NUM_LAUNCHED_JOBS=$((${NUM_LAUNCHED_JOBS} + 1))
	WaitUntilReadyToQueuePID
done

wait

for BENCH in ${BENCHS}; do
	printf "${BENCH}:" | tee -a log.txt

	if [ -n "$(cat sim_out_${BENCH}.txt | grep 'ERROR!')" ]; then
		printf "FAIL (see sim_out_${BENCH}.txt)\n"
	else
		cat console_out_${BENCH}.txt \
		| sed -n "/Starting MiBench/,/MiBench .* is finished/p" \
		| sed '1d' | sed '$d' | sed 's/\x0D$//' \
		| sed -e 's/^\(.*\)Time:   [0-9][0.9]*\.[0-9][0-9]* sec\(.*\)$/\1Time:    \*.\*\*\* sec\2/g' -e 's///g' -e 's/^Best  >.*$/Best  > N\/A/' -e 's/^Worst >.*$/Worst > N\/A/' \
		| sed -e 's/^%%CreationDate:.*$/%%CreationDate: Thu Jan  1 00:00:00 1970/' \
		> output_${BENCH}.txt
		diff "output_${BENCH}.ref.txt" "output_${BENCH}.txt" &> /dev/null
		if [ $? -eq 0 ]; then
			printf "PASS\n" | tee -a log.txt
		else
			printf "FAIL (see console_out_${BENCH}.txt)\n" | tee -a log.txt
		fi
	fi
done
